structure LambdaPlusPlus =
struct
  datatype typ =
    BaseType of string
  | Arrow of typ * typ
  | Plus of typ * typ
  | Times of typ * typ

  datatype exp =
    Variable of string
  | Lambda of string * exp
  | Apply of exp * (exp * typ)
  | Tuple of exp * exp
  | First of exp * typ
  | Second of exp * typ
  | Left of exp
  | Right of exp
  | Case of exp * ((string * typ) * exp) * ((string * typ) * exp)

  infix 6 ^^
  fun s1 ^^ s2 = s1 ^ " " ^ s2
  val rec pp_typ = (fn
      BaseType a => a
    | Arrow (t1, t2) => "(" ^ pp_typ t1 ^^ "->" ^^ pp_typ t2 ^ ")"
    | Plus (t1, t2) => "(" ^ pp_typ t1 ^^ "+" ^^ pp_typ t2 ^ ")"
    | Times (t1, t2) => "(" ^ pp_typ t1 ^^ "*" ^^ pp_typ t2 ^ ")"
  )

  val rec pp_exp = (fn
      Variable x => x
    | Lambda (x, e) => "fn" ^^ x ^^ "=>" ^^ pp_exp e
    | Apply (e1, (e2, t2)) => "(" ^ pp_exp e1 ^^ "(" ^ pp_exp e2 ^^ ":" ^^ pp_typ t2 ^ "))"
    | Tuple (e1, e2) => "(" ^ pp_exp e1 ^ "," ^^ pp_exp e2 ^ ")"
    | First (e, t) => "#1 (" ^ pp_exp e ^^ ":" ^^ pp_typ t ^ ")"
    | Second (e, t) => "#2 (" ^ pp_exp e ^^ ":" ^^ pp_typ t ^ ")"
    | Left e => "INL" ^^ pp_exp e
    | Right e => "INR" ^^ pp_exp e
    | Case (e, ((x1, t1), e1), ((x2, t2), e2)) =>
      "case" ^^ pp_exp e ^^ "of" ^^
      "INL" ^^ x1 ^^ ":" ^^ pp_typ t1 ^^ "=>" ^^ pp_exp e1 ^^ "|" ^^
      "INR" ^^ x2 ^^ ":" ^^ pp_typ t2 ^^ "=>" ^^ pp_exp e2
  )

  structure Parser =
  struct
    val Left_ = Left
    val Right_ = Right
    open ParserCombinators
    open CharParser

    infixr 4 << >>
    infixr 3 &&
    infix  2 -- ##
    infix  2 wth suchthat return guard when
               infixr 1 || <|> ??

    structure LPPDef :> LANGUAGE_DEF =
    struct

      type scanner = char CharParser.charParser

      val commentStart   = SOME "(*"
      val commentEnd     = SOME "*)"
      val commentLine    = NONE
      val nestedComments = true

      val identStart     = CharParser.letter
      val identLetter    = CharParser.alphaNum
      val opStart        = CharParser.oneOf [#"=", #"#", #"|", #",", #"-", #"+", #"*", #":"]
      val opLetter       = CharParser.oneOf [#">", #"1", #"2"]
      val reservedNames  = ["fn", "INL", "INR", "case", "of"]
      val reservedOpNames= ["=>", "|", "#1", "#2", ",", "->", "+", "*", ":"]
      val caseSensitive  = true

    end

    structure LPP = TokenParser (LPPDef)
    open LPP

    val tvar = identifier wth BaseType
    fun arrowify [] = raise Fail "you are tearing me apart"
      | arrowify [t] = t
      | arrowify (t1::t2::ts) = Arrow (t1, arrowify (t2::ts))
    fun typ () = separate1 ($typ') (reservedOp "->") wth arrowify
    and typ' () =
          ($typ'' -- (fn t1 => reservedOp "+" >> $typ'' wth (fn t2 => Plus (t1, t2))))
      ||  $typ''
    and typ'' () =
          ($typ''' -- (fn t1 => reservedOp "*" >> $typ''' wth (fn t2 => Times (t1, t2))))
      ||  $typ'''
    and typ''' () = tvar <|> parens ($typ)

    val full_type = $typ << (eos ?? "end of stream")

    val var = identifier wth Variable
    fun annotated_identifier () =
          parens ($annotated_identifier)
      <|> (identifier -- (fn x => reservedOp ":" >> $typ wth (fn t => (x, t))))
    and annotated_exp () =
          parens ($annotated_exp)
      <|> ($exp -- (fn e => reservedOp ":" >> $typ wth (fn t => (e, t))))
    and exp () =
      $exp' -- (fn e => repeat (parens ($annotated_exp)) wth (
        foldl (fn (arg, f) => Apply (f, arg)) e
      ))
    and exp' () =
      parens ($exp) ||
          var
      <|> $lambda
      <|> $tuple
      <|> $first
      <|> $second
      <|> $left
      <|> $right
      <|> $case_analysis
    and apply () =
      $exp -- (fn e1 => parens ($annotated_exp) wth (fn et2 => Apply (e1, et2)))
    and lambda () =
      reserved "fn" >>
      identifier -- (fn x =>
        reservedOp "=>" >>
        $exp wth (fn e =>
          Lambda (x, e)
        )
      )
    and tuple () = parens (
        $exp -- (fn e1 =>
          reservedOp "," >>
          $exp wth (fn e2 =>
            Tuple (e1, e2)
          )
        )
      )
    and first () = reservedOp "#1" >> $annotated_exp wth First
    and second () = reservedOp "#2" >> $annotated_exp wth Second
    and left () = reserved "INL" >> $exp wth Left_
    and right () = reserved "INR" >> $exp wth Right_
    and case_analysis () =
      reserved "case" >>
      $exp -- (fn e =>
        reserved "of" >>
        reserved "INL" >>
        $annotated_identifier -- (fn xt1 =>
          reservedOp "=>" >>
          $exp -- (fn e1 =>
            reservedOp "|" >>
            reserved "INR" >>
            $annotated_identifier -- (fn xt2 =>
              reservedOp "=>" >>
              $exp wth (fn e2 =>
                Case (e, (xt1, e1), (xt2, e2))
              )
            )
          )
        )
      )
    val program = $exp << (eos ?? "end of stream")
  end

  val parse_exp = CharParser.parseString Parser.program
  val parse_typ = CharParser.parseString Parser.full_type
end
